package org.antlr.tool;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.antlr.Tool;
import org.antlr.analysis.DFAState;
import org.antlr.analysis.NFAConfiguration;
import org.antlr.analysis.NFAState;
import org.antlr.analysis.RuleClosureTransition;
import org.antlr.analysis.SemanticContext;
import org.antlr.analysis.State;
import org.antlr.analysis.Transition;
import org.antlr.misc.Utils;
import org.antlr.tool.Grammar;
import org.antlr.tool.GrammarAST;
import org.stringtemplate.v4.ST;
import org.stringtemplate.v4.STGroup;
import org.stringtemplate.v4.STGroupFile;

public class DOTGenerator {
  public static final boolean STRIP_NONREDUCED_STATES = false;
  
  protected String arrowhead = "normal";
  
  protected String rankdir = "LR";
  
  public static STGroup stlib = new STGroupFile("org/antlr/tool/templates/dot/dot.stg");
  
  protected Set<Object> markedStates = null;
  
  protected Grammar grammar;
  
  public DOTGenerator(Grammar grammar) {
    this.grammar = grammar;
  }
  
  public String getDOT(State startState) {
    ST dot;
    if (startState == null)
      return null; 
    this.markedStates = new HashSet();
    if (startState instanceof DFAState) {
      dot = stlib.getInstanceOf("dfa");
      dot.add("startState", Utils.integer(startState.stateNumber));
      dot.add("useBox", Boolean.valueOf(Tool.internalOption_ShowNFAConfigsInDFA));
      walkCreatingDFADOT(dot, (DFAState)startState);
    } else {
      dot = stlib.getInstanceOf("nfa");
      dot.add("startState", Utils.integer(startState.stateNumber));
      walkRuleNFACreatingDOT(dot, startState);
    } 
    dot.add("rankdir", this.rankdir);
    return dot.render();
  }
  
  protected void walkCreatingDFADOT(ST dot, DFAState s) {
    ST st;
    if (this.markedStates.contains(Utils.integer(s.stateNumber)))
      return; 
    this.markedStates.add(Utils.integer(s.stateNumber));
    if (s.isAcceptState()) {
      st = stlib.getInstanceOf("stopstate");
    } else {
      st = stlib.getInstanceOf("state");
    } 
    st.add("name", getStateLabel(s));
    dot.add("states", st);
    for (int i = 0; i < s.getNumberOfTransitions(); i++) {
      Transition edge = s.transition(i);
      st = stlib.getInstanceOf("edge");
      st.add("label", getEdgeLabel(edge));
      st.add("src", getStateLabel(s));
      st.add("target", getStateLabel(edge.target));
      st.add("arrowhead", this.arrowhead);
      dot.add("edges", st);
      walkCreatingDFADOT(dot, (DFAState)edge.target);
    } 
  }
  
  protected void walkRuleNFACreatingDOT(ST dot, State s) {
    ST stateST;
    if (this.markedStates.contains(s))
      return; 
    this.markedStates.add(s);
    if (s.isAcceptState()) {
      stateST = stlib.getInstanceOf("stopstate");
    } else {
      stateST = stlib.getInstanceOf("state");
    } 
    stateST.add("name", getStateLabel(s));
    dot.add("states", stateST);
    if (s.isAcceptState())
      return; 
    if (((NFAState)s).isDecisionState()) {
      GrammarAST n = ((NFAState)s).associatedASTNode;
      if (n != null && n.getType() != 33) {
        ST rankST = stlib.getInstanceOf("decision-rank");
        NFAState alt = (NFAState)s;
        while (alt != null) {
          rankST.add("states", getStateLabel(alt));
          if (alt.transition[1] != null) {
            alt = (NFAState)(alt.transition[1]).target;
            continue;
          } 
          alt = null;
        } 
        dot.add("decisionRanks", rankST);
      } 
    } 
    for (int i = 0; i < s.getNumberOfTransitions(); i++) {
      Transition edge = s.transition(i);
      if (edge instanceof RuleClosureTransition) {
        RuleClosureTransition rr = (RuleClosureTransition)edge;
        ST edgeST = stlib.getInstanceOf("edge");
        if (rr.rule.grammar != this.grammar) {
          edgeST.add("label", "<" + rr.rule.grammar.name + "." + rr.rule.name + ">");
        } else {
          edgeST.add("label", "<" + rr.rule.name + ">");
        } 
        edgeST.add("src", getStateLabel(s));
        edgeST.add("target", getStateLabel(rr.followState));
        edgeST.add("arrowhead", this.arrowhead);
        dot.add("edges", edgeST);
        walkRuleNFACreatingDOT(dot, rr.followState);
      } else {
        ST edgeST;
        if (edge.isAction()) {
          edgeST = stlib.getInstanceOf("action-edge");
        } else if (edge.isEpsilon()) {
          edgeST = stlib.getInstanceOf("epsilon-edge");
        } else {
          edgeST = stlib.getInstanceOf("edge");
        } 
        edgeST.add("label", getEdgeLabel(edge));
        edgeST.add("src", getStateLabel(s));
        edgeST.add("target", getStateLabel(edge.target));
        edgeST.add("arrowhead", this.arrowhead);
        dot.add("edges", edgeST);
        walkRuleNFACreatingDOT(dot, edge.target);
      } 
    } 
  }
  
  protected String getEdgeLabel(Transition edge) {
    String label = edge.label.toString(this.grammar);
    label = Utils.replace(label, "\\", "\\\\");
    label = Utils.replace(label, "\"", "\\\"");
    label = Utils.replace(label, "\n", "\\\\n");
    label = Utils.replace(label, "\r", "");
    if (label.equals("<EPSILON>"))
      label = "e"; 
    State target = edge.target;
    if (!edge.isSemanticPredicate() && target instanceof DFAState) {
      SemanticContext preds = ((DFAState)target).getGatedPredicatesInNFAConfigurations();
      if (preds != null) {
        String predsStr = "&&{" + preds.genExpr(this.grammar.generator, this.grammar.generator.getTemplates(), null).render() + "}?";
        label = label + predsStr;
      } 
    } 
    return label;
  }
  
  protected String getStateLabel(State s) {
    if (s == null)
      return "null"; 
    String stateLabel = String.valueOf(s.stateNumber);
    if (s instanceof DFAState) {
      StringBuilder buf = new StringBuilder(250);
      buf.append('s');
      buf.append(s.stateNumber);
      if (Tool.internalOption_ShowNFAConfigsInDFA) {
        if (s instanceof DFAState && 
          ((DFAState)s).abortedDueToRecursionOverflow) {
          buf.append("\\n");
          buf.append("abortedDueToRecursionOverflow");
        } 
        Set<Integer> alts = ((DFAState)s).getAltSet();
        if (alts != null) {
          buf.append("\\n");
          List<Integer> altList = new ArrayList<Integer>();
          altList.addAll(alts);
          Collections.sort(altList);
          Set<NFAConfiguration> configurations = ((DFAState)s).nfaConfigurations;
          for (int altIndex = 0; altIndex < altList.size(); altIndex++) {
            Integer altI = altList.get(altIndex);
            int alt = altI.intValue();
            if (altIndex > 0)
              buf.append("\\n"); 
            buf.append("alt");
            buf.append(alt);
            buf.append(':');
            List<NFAConfiguration> configsInAlt = new ArrayList<NFAConfiguration>();
            for (NFAConfiguration c : configurations) {
              if (c.alt != alt)
                continue; 
              configsInAlt.add(c);
            } 
            int n = 0;
            for (int cIndex = 0; cIndex < configsInAlt.size(); cIndex++) {
              NFAConfiguration c = configsInAlt.get(cIndex);
              n++;
              buf.append(c.toString(false));
              if (cIndex + 1 < configsInAlt.size())
                buf.append(", "); 
              if (n % 5 == 0 && configsInAlt.size() - cIndex > 3)
                buf.append("\\n"); 
            } 
          } 
        } 
      } 
      stateLabel = buf.toString();
    } 
    if (s instanceof NFAState && ((NFAState)s).isDecisionState()) {
      stateLabel = stateLabel + ",d=" + ((NFAState)s).getDecisionNumber();
      if (((NFAState)s).endOfBlockStateNumber != -1)
        stateLabel = stateLabel + ",eob=" + ((NFAState)s).endOfBlockStateNumber; 
    } else if (s instanceof NFAState && ((NFAState)s).endOfBlockStateNumber != -1) {
      NFAState n = (NFAState)s;
      stateLabel = stateLabel + ",eob=" + n.endOfBlockStateNumber;
    } else if (s instanceof DFAState && ((DFAState)s).isAcceptState()) {
      stateLabel = stateLabel + "=>" + ((DFAState)s).getUniquelyPredictedAlt();
    } 
    return '"' + stateLabel + '"';
  }
  
  public String getArrowheadType() {
    return this.arrowhead;
  }
  
  public void setArrowheadType(String arrowhead) {
    this.arrowhead = arrowhead;
  }
  
  public String getRankdir() {
    return this.rankdir;
  }
  
  public void setRankdir(String rankdir) {
    this.rankdir = rankdir;
  }
}
